// Copyright 1999-2005 Jon Dart. All Rights Reserved.

#include "hash.h"
#include "util.h"
#include "debug.h"
#include "constant.h"
#include "globals.h"
#include "legal.h"
#include "learn.h"
extern "C" {
#ifndef _MAC
#include <malloc.h>
#endif
#include <memory.h>
#include <stddef.h>
};

static DualHashEntry *hashTable = NULL;
int hashSize;
static const int MaxRehash = 3;
static int refCount = 0;
static int hash_init_done = 0;

void initHash(uint64 bytes)
{
    int hash_entries;
    hash_entries = (int)(bytes/sizeof(HashEntry));
    if (!hash_init_done) {
       hashSize = hash_entries/2 + (hash_entries/2 % 2);  // ensure it's even
       ALIGNED_MALLOC(hashTable,
         DualHashEntry,
         sizeof(DualHashEntry)*hashSize,128);
       clearHash();
       hash_init_done++;
       ++refCount;
    }
}

void resizeHash(uint64 bytes) {
  freeHash();
  initHash(bytes);      
}

void freeHash() {
    if (--refCount == 0) {
        ALIGNED_FREE(hashTable);
        hash_init_done = 0;
    }
}


void clearHash() {
    memset(hashTable,'\0',sizeof(DualHashEntry)*hashSize);
}

void loadLearnInfo() {
  if (options.learning.position_learning) {
    ifstream plog;
    plog.open(learnFileName.c_str(),ios_base::in);
    while (plog.good() && !plog.eof()) {
      LearnRecord rec;
      if (getLearnRecord(plog,rec)) {
	 Move best = NullMove;
	 if (rec.start != InvalidSquare) 
	   best = CreateMove(rec.start,rec.dest,rec.promotion);
         HashEntry pos(rec.hashcode,rec.depth*DEPTH_INCREMENT,
                   0, /* age */
                   PositionInfo::Valid,
                   rec.in_check,
	           rec.score,1,
                   best);
	 storeHash(pos);
      }
    }
  }
}

PositionInfo::ValueType searchHash(const Board& b,hash_t hashCode,
int alpha, int beta, int ply,
int depth,
PositionInfo &entry
)
{
    int i;
    int probe = quickmod(hashCode,hashSize);
    DualHashEntry *p = (DualHashEntry*)&(hashTable[probe]);
    HashEntry *hit = NULL;
    if (p->replace.hash_code() == hashCode) {
        entry.copy(p->replace);
        hit = &p->replace;
        if (p->replace.depth() >= depth) {
            // usable depth
            goto hash_hit;
        }
        // got a hit but depth not sufficient: keep the
        // data, but try the depth table too
    }
    for (i = MaxRehash; i != 0; --i) {
        if (p->depth.hash_code() == hashCode) {
            entry.copy(p->depth);
            hit = &p->depth;
            if (p->depth.depth() >= depth) {
                // usable depth
                goto hash_hit;
            }
            else {
                break;
            }
        }
        probe++; p++;
        if (probe >= hashSize) {
            probe = 0; p = hashTable;
        }
    }
    // If we got a hash hit but insufficient depth return here:
    if (hit) return PositionInfo::Invalid;
    return PositionInfo::NoHit;
hash_hit:
    //
    // Whether we can use the hash value or not depends on
    // the flags:
    //
    int value = entry.value();
    switch (entry.type()) {
        case PositionInfo::Valid:
            // If this is a mate score, adjust it to reflect the
            // current ply depth.
            //
            if (value >= Constants::MATE-256) {
                value -= ply;
                entry.setValue(value);
            }
            else if (value <= -Constants::MATE+256) {
                value += ply;
                entry.setValue(value);
            }
            return PositionInfo::Valid;
        case PositionInfo::LowerBound:
            if (value >= beta)
                return PositionInfo::LowerBound;
            break;
        case PositionInfo::UpperBound:
            if (value <= alpha)
                return PositionInfo::UpperBound;
            break;
        default:
            break;
    }                                             // end switch
    return PositionInfo::Invalid;
}


void storeHash(const HashEntry &newPos,Statistics *stats) {
    int probe = quickmod(newPos.hash_code(),hashSize);
    DualHashEntry *p = (DualHashEntry*)&(hashTable[probe]);
    p->replace = newPos;

    HashEntry *best = NULL;
    int minDepth = 10000;
    ASSERT(Util::Abs(newPos.value()) <= Constants::MATE);
    // Of the positions that hash to the same locations
    // as this one, find the best one to replace.
    for (int i = MaxRehash; i != 0; --i) {
        HashEntry *q = &p->depth;

        if (q->hash_code() == (hash_t)0 || 
	    (q->age() && (q->age() != newPos.age()))) {
            *q = newPos; 
	    if (stats) {
               stats->hash_inserts++;
               stats->new_hash_inserts++;
	    }
            return;  // stored
        }
        if (q->hash_code() == newPos.hash_code()) {
            // same position
            q->update_best_move(newPos);
            if (q->depth() <= newPos.depth()) {
               best = q;
               break;
           }
        }
        if (q->depth() < minDepth && q->age()) {
            minDepth = q->depth();
            if (newPos.depth() >= minDepth) best = q;
        }
        probe++; p++;
        if (probe >= hashSize) {
            probe = 0; p = hashTable;
        }
    }
    if (best != NULL) {
        //if (best->age() != newPos.age() && stats)
        //  stats->new_hash_inserts++;
        *best = newPos;
        if (stats) stats->hash_replaces++;
        return;
    }
    else if (stats) {
      stats->hash_inserts_failed++;
    }
}

Move PositionInfo::best_move(const Board &b) const {
    if (start == InvalidSquare)
         return NullMove;
    else {
       Move m = CreateMove(b,(Square)start,(Square)dest,
          (PieceType)promotion);
       if (!validMove(b,m)) m = NullMove;
       if (forced()) SetForced(m);
       return m;
    }
}

